package de.tub.mapc.components.contestagents;

import java.io.IOException;
import java.util.Enumeration;

import org.slf4j.Logger;

import de.dailab.keks.io.IAddress;
import de.dailab.keks.io.IMessage;
import de.dailab.keks.io.IMulticastAddress;
import de.jiac.micro.agent.AbstractReactiveBehaviour;
import de.jiac.micro.agent.handle.ICommunicationHandle;
import de.jiac.micro.agent.memory.ILongTermMemory;
import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.feature.ILifecycleAware;
import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.monitor.AgentMonitor;

public class SimulationStartElement extends AbstractReactiveBehaviour implements ILifecycleAware{

	private Logger _logger;
	private IAgent agent;
	private ICommunicationHandle _communicationHandle;
	private IMulticastAddress _address;
//	private IAddress monitoragentaddress;
	private IAddress coordinatoraddress;
	private ILongTermMemory memory;
	private AgentMonitor monitor;
	
	@SuppressWarnings("unchecked")
	@Override
	protected Class[] filterDataTypes() {
		return new Class[] {Gridworld.class};
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void runShort(Enumeration sensorReadings) {
		while (sensorReadings.hasMoreElements()) {
			Object read = sensorReadings.nextElement();
			if (read instanceof Gridworld){
				handleGridworld((Gridworld)read);
			}
		}
	}
	
	/**
	 * 
	 * @param fact
	 */
	private void handleGridworld(Gridworld fact){
		// access longTermMemory
		// and save the grid
		memory.memorise(fact);
		
		monitor.init(fact);
		
		// tell the coordinator (the other agents got the same message, they dont need it) 
		IMessage message = _communicationHandle.createMessage();
		message.setContent(fact);
		try {
			_communicationHandle.sendMessage(coordinatoraddress, message);
		} catch (Exception e) {
			_logger.error("Could not send the message to the coordinatoragent", e);
		}
	}
	
	
	@Override
	public void initialise(){
		super.initialise();
		agent = AgentScope.getAgentReference();
		// create the pathFinder
		IPathFindingHandle pathfinder = new PathFindingElement();
		agent.addHandle(pathfinder);
		
	}

	@Override
	public void start() {
        super.start();
        
		// Referenz auf den Agenten
		agent = AgentScope.getAgentReference();
		// Referenz auf den Logger
		_logger = agent.getLogger("SimulationStart");
		memory = (ILongTermMemory) agent.getHandle(ILongTermMemory.class);
		// Referenz auf die Kommunikationsschnittstelle
		_communicationHandle = (ICommunicationHandle) agent.getHandle(ICommunicationHandle.class);
		
		// Referenz auf die Monitor
		monitor = AgentMonitor.getInstance();
		if (monitor == null) {
			_logger.error("SSE: Monitor not initialised");
		}
		
		// set up the communication with other agents
		try {
			// use the teamname as groupname for communication
			String groupName = (String) agent.getProperty(IMAPCConstants.TEAMNAME_KEY);
			_address = _communicationHandle.getMulticastAddressForName(groupName);
			_communicationHandle.joinGroup(_address);
			
			coordinatoraddress= _communicationHandle.getMulticastAddressForName("coordinator");

		} catch (IOException e) {
			_logger.error("Could not join the team group", e);
		}
		
	}

	@Override
	public void stop() {

		// Der Agent faehrt runter -> wir brauchen die Gruppe nicht mehr.
		try {
			_communicationHandle.leaveGroup(_address);
		} catch (IOException e) {
			_logger.error("Could not leave the team group", e);
		}
		
		super.stop();
	}
	
	
	
}
