package de.tub.mapc.components.coordinatoragent;

import java.awt.Point;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.slf4j.Logger;

import de.dailab.keks.io.IAddress;
import de.dailab.keks.io.IMessage;
import de.dailab.keks.io.IMulticastAddress;
import de.jiac.micro.agent.AbstractReactiveBehaviour;
import de.jiac.micro.agent.handle.ICommunicationHandle;
import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.feature.ILifecycleAware;
import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.Coordination.Agent;
import de.tub.mapc.Coordination.CoordinationData;
import de.tub.mapc.Coordination.ExploringPoint;
import de.tub.mapc.Coordination.Voter;
import de.tub.mapc.Coordination.goals.AbstractGoal;
import de.tub.mapc.Coordination.goals.GoalBlockEnemyCorral;
import de.tub.mapc.Coordination.goals.GoalDriveHerd;
import de.tub.mapc.Coordination.goals.GoalExploreField;
import de.tub.mapc.Coordination.goals.GoalFindSwitchForFence;
import de.tub.mapc.Coordination.goals.GoalGoToSwitchingPosition;
import de.tub.mapc.Coordination.messages.MessageSwitchVote;
import de.tub.mapc.Coordination.messages.SwitchAnfrage;
import de.tub.mapc.Coordination.messages.SwitchingAnswer;
import de.tub.mapc.components.contestagents.PathFindingElement;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.lang.IntHashMap;
import de.tub.mapc.monitor.AgentMonitor;
import de.tub.mapc.pathfinding.PathResult;
import de.tub.mapc.worldmodel.Herd;
import de.tub.mapc.worldmodel.HerdOrganisation;

public class CoordinationElement extends AbstractReactiveBehaviour implements
		ILifecycleAware {
    /**
     * number of agents that have to vote for a switch, before the coordinator
     * sends a switcher<br/>
     * eventuell andere Mindestgrenze festlegen, je mehr noetig
     * sind, desto mehr gehen auch durch, wichtig wenn mehrere zaeune
     * hintereinander andererseits wenn jemand kuehe durchtreibt will er nicht
     * ewig warten. leute mit kuehen hoeher gewichten... ACHTUNG, PRIO DER GOALS
     * BEACHTEN.... NICHT MEHR ANZAHL DER VOTES Prio: - explore 10 - switch 20 -
     * herde erreichen 30 - treiben 40
     */
    private static final int MINIMUM_VOTERS = 20;
    /**
     * number of agents, only these ones which have an connection to the
     * contestserver, not the coordinator, monitor or something like this
     */
    private static final int NUMBER_OF_AGENTS = 20;
    private static final int MIN_EXPLORER = 2;
    
    private static final int NUM_ROUNDS_BETWEEN_HERD_UPDATES = 0; //rounds between update
    private static final int NUM_ROUNDS_BETWEEN_HERD_REBUILDS = 10; //rounds between rebuild
    private static final int NUM_ROUNDS_TO_FASTER_REBUILD = NUM_ROUNDS_BETWEEN_HERD_REBUILDS / 3; //rounds between next rebuild (for faster rebuild if we have no herds); 

	// TODO eventuell waere es sinnvoll den switch beim corral immer besetzt zu
	// lassen
	// also so dass er 2 Felder daneben stehen bleibt, denn diesmal ist es
	// wichtig, dass
	// mittendrin kuehe schnell reinkoennen und dann auch drin bleiben
	// es wird ja der Durchschnitt gewertet.

	private int updateHerdsCounter = 0;
	private int rebuildHerdsCounter = 0;

	private ICommunicationHandle _communicationHandle;
	private Logger _logger;
	private IMulticastAddress _address;
	private CoordinationData data;
	private HerdOrganisation herdOrga; 
	
	@SuppressWarnings("unchecked")
	@Override
	protected Class[] filterDataTypes() {
		// the new information comes from messages sent by other agents
		return new Class[] { IMessage.class };
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void runShort(Enumeration sensorReadings) {
		while (sensorReadings.hasMoreElements()) {
			IMessage message = (IMessage) sensorReadings.nextElement();
			Object content = message.getContent();
			String sourceStr= message.getHeader(IMessage.DefaultHeader.SOURCE_ADDRESS);
			IAddress sender= _communicationHandle.getAddressForString(sourceStr);
			if (content instanceof Perception) {
				handlePerception((Perception) content, sender);
			} else if (content instanceof Gridworld) {
				// a new world received, initialize worldmodel
				initializeData((Gridworld) content);
			} else if (content instanceof MessageSwitchVote) {
				handleSwitchVote((MessageSwitchVote) content, data.getAgent(sender));
			} else if (content instanceof SwitchingAnswer) {
				handleSwitchingAnswer((SwitchingAnswer) content, sourceStr);
			} else if (content instanceof String) {
				// a textmessage received, analyze the string
				String s = (String) content;
				if (s.equals(IMAPCConstants.MESSAGE_RESET)) {
					// we have to prepare for a new simulation
					data.reset();
				}
			}
		}
	}

	private void handleSwitchVote(MessageSwitchVote vote, Voter voter) {
		
		if (vote.isFence()) {
			// position of the switch is not known, save vote for exploring the
			// switch
			List<Switch> unknownSwitches = data.getUnknownSwitches();
			boolean found = false;
			// fuer jeden pruefen ob, x oder y gleich ist, dann ist es
			// (wahrscheinlich)
			// der gleiche Zaun
			// wenn nix gefunden, dann ist keiner dahin unterwegs und der vote
			// muss eingetragen werden
			for (Switch alreadyVoted : unknownSwitches) {
				if (alreadyVoted.getId() == vote.getId()) {

					// there is already a vote for this fencefield or
					// the vote is in the same row, probably it is the same
					// fence
					// -> add vote on the existing entry
					alreadyVoted.addVoter(voter, vote.getDistance());
					// _logger.trace("COORD: there are already votes for this fence to explore("+alreadyVoted.toString()+"), adding one");
					found = true;
					break;
				}
			}
			if (!found) {
				// _logger.trace("COORD: this was the first vote for this fence to explore");
				// add a new Entry to the list
				Switch s = new Switch(vote.getX(), vote.getY());
				s.addVoter(voter, vote.getDistance());
				data.addUnknownSwitch(s);
			}
		} else {
			// save vote in List
			Switch s = data.getSwitches().get(vote.getId());
			s.addVoter(voter, vote.getDistance());
		}
	}
	

	private void handleSwitchingAnswer(SwitchingAnswer ans, String sender) {

		// FIRST STEP: SAVE ANSWERS
		Switch s = data.getSwitches().get(ans.getId());
		// _logger.trace("SWITCHING ANSWER COUNT: " +
		// (s.getNotReachingAgents().size()+s.getPossibleSwitchers().size()+1));
		if (ans.getDistance() == -1) {
			s.addnonReachingAgent(sender);
		} else {
			// _logger.trace("COORD RECEIVED POSITIVE SWITCHING ANSWER: "+ans.getDistance());
			// maybe there were enough answers and we have already chosen an
			// agent
			if (s.getSwitchingAgent().equals("")) {
				// no, there is no switching agent
				ans.setSender(sender);
				s.addPossibleSwitcher(ans);
			} else {
				// _logger.info("COORD: THERE IS ALREADY A SWITCHER, IGNORING SWITCH ANSWER");
			}
		}
		// SECOND STEP: WHEN ALL ANSWERS REACHED
		if ((s.getPossibleSwitchers().size() + s.getNotReachingAgents().size() == (s
				.getRequestedAgents().size()))
				&& (!s.getPossibleSwitchers().isEmpty())) {

			// TODO prioritaet der goals beachten? jetzt koennte ein wichtiger
			// fuer weniger wichtige den switcher spielen

			// get the best answer and save it in list, so the goal will be
			// assigned to the agent in next step
			data.addSwitchToBeAssignedFromRequest(s.getPossibleSwitchers()
					.poll());
			// the voting Data for this switch can be reset
			s.clear();
		} 
		// TODO was wenn keiner erreichen kann?
	}

	// TODO wenn das noch weiter waechst zerlegen, zb den Bereich, wo alle Perceptions da sind
	private void handlePerception(Perception perception, IAddress sender) {
		if (data == null)
			return;
		if (perception.getStep() > data.getStepNo()) {
			_logger.trace("##################################################");
			// New step, set the received perceptions to 0
			data.resetReceivedPerceptions();
			data.setStepNo(perception.getStep());
		}

		// a perception received, update worldmodel
		// this perception will be counted in data
		updateData(perception, sender);
		
		// TODO reachedGoal aus Perception speichern
		

		// TODO can the number be changed if an agent loses connection or
		// something like this?
		if (data.getReceivedPerceptions() == NUMBER_OF_AGENTS) {
			// set the votes in the switches one step older
			Switch[] switches = data.getSwitches().valuesAsArray(Switch.class);
			for (Switch s : switches) {
				s.getVotings().step();
			}
			// the same for unknown switches
			List<Switch> unknownSwitches = data.getUnknownSwitches();
			for (Switch s : unknownSwitches) {
				s.getVotings().step();
			}
			AgentMonitor.getInstance().setWorld(data.getWorld(),perception.getStep(),perception.getScore());
			//Welt der Herden aktualisieren
			if(herdOrga!=null)
				herdOrga.updateWorldOfHerds(data.getWorld());
			assignGoals();
			sendGoals();
		}
	}

	/**
	 * sends the current goal, which is stored in the agent himself, to every agent
	 */
	private void sendGoals() {
		HashMap<String, Agent> agents = data.getAgents();
		Set<String> keys = agents.keySet();
		for (String key : keys) {
			this.sendGoalToAgent(agents.get(key));
		}
	}
	
	//TODO in kleine Teile zerlegen, aber erst wenn es fertig ist!
	private void assignGoals() {
		List<Agent> uncheckedAgents = new ArrayList<Agent>();
		uncheckedAgents.addAll(data.getAgents().values());

	//--------------SWITCHING--------------------------------------
		//TODO sinnvoll?: switchGoals sind jetzt so viele runden gueltig, wie die max. distanz eines voters zum zaun war
		
		// die zuordnung von switchrequest uebernehmen
		List<SwitchingAnswer> requestetSwitches = data
				.getSwitchesToBeAssignedFromRequest();
		// _logger.info("COORD: NUMBER OF SWITCHES THAT NEED TO BE ASSIGNED FROM REQUEST: "+requestetSwitches.size());
		Iterator<SwitchingAnswer> i = requestetSwitches.iterator();
		SwitchingAnswer currentSwitchingAnswer;
		int rounds;
		while (i.hasNext()) {
			currentSwitchingAnswer = i.next();
			Switch s = data.getSwitches().get(currentSwitchingAnswer.getId());
			Agent a = data.getAgent(currentSwitchingAnswer.getSender());
			s.setSwitchingAgent(currentSwitchingAnswer.getSender());
			rounds=getRoundsForSwitchGoal(s);
			this.setNewGoal(a, new GoalGoToSwitchingPosition(currentSwitchingAnswer
					.getSwitchingPosition(), s, rounds));
			uncheckedAgents.remove(a);
			// _logger.info("COORD: SET AGENT TO SWITCH FROM REQUEST");
			// remove the switchingAnswer from list
			i.remove();
		}

		// wenn es einen Vote ohne bekannten Zaun gab,
		// dann Zaunexplorer einteilen
		List<Switch> unknownSwitches = data.getUnknownSwitches();
		for (Switch currentUnknownSwitch : unknownSwitches) {
			if (currentUnknownSwitch.getVotings().getVotingCount() >= MINIMUM_VOTERS) {
				if (currentUnknownSwitch.getSwitchingAgent() != "") {
					// switcher(switchsearcher) keeps his goal
					uncheckedAgents.remove(data.getAgent(currentUnknownSwitch
							.getSwitchingAgent()));
					// _logger.trace("schon einer unterwegs um Switch bei "+currentUnknownSwitch
					// +" zu erkunden");
					continue;
				}
				// _logger.trace("versuchen einen zum erkunden von Switch bei "+currentUnknownSwitch+" zu schicken");
				// one of the voters will be switchsearcher
				List<Agent> votingAgents = currentUnknownSwitch.getVotings()
						.getVotingAgents();
				boolean foundSwitchExplorer = false;
				for (Agent possibleSwitchExplorer : votingAgents) {
					// check, wheter this voter can reach the fence
					// TODO jetzt wird es der erste in der Liste, welcher
					// erreichen kann
					// der am naechsten dran waer besser
					// wenn der nicht erreicht, is der zweite dran usw.
					// ODER EINFACH BEIM VOTE DIE ANZAHL DER SCHRITTE BIS ZUM
					// ZAUN AUSLESEN UND MITSCHICKEN UND DAVON DIE GERINGSTE
					// NEHMEN
					// (GLEICH SORTIERT IN DIE VOTINGLISTE REINPACKEN ODER SO)
					
					// GIBTS SCHON BEI NORMALEN VOTES
					
					IPathFindingHandle pathfinder = new PathFindingElement(false);
					PathResult pr = pathfinder.getPath(currentUnknownSwitch,
							possibleSwitchExplorer.getPosition(), data.getWorld(), false);
					if ((pr.getErrortype() == IMAPCConstants.PATH_FENCEAHEAD)
							&& (pr.getPathLength() == pr.getStepsToNextFence())) {
						foundSwitchExplorer = true;
						// this agent can be switchexplorer, assign goal
						this.setNewGoal(possibleSwitchExplorer,	new GoalFindSwitchForFence(currentUnknownSwitch));
						uncheckedAgents.remove(possibleSwitchExplorer);
						currentUnknownSwitch.setSwitchingAgent(possibleSwitchExplorer.getAddress().toString());
						// and do not check the other ones
						break;
					}

				}
				if (!foundSwitchExplorer) {
					// TODO if no voter can reach, end switch request to every
					// agent
					// TODO waer besser nicht zu jedem, voter brauchen nicht
					// mehr berechnet werden, die wurden ja schon geprueft
					// das hier ist sehr rechenintensiv, weil zu jeder switching
					// position ein pfad berechnet wird
					
					// DAS GIBTS SCHON BEI NORMALEN SWITCH REQUESTS, KANN VON DA UEBERNOMMEN WERDEN
				}
			} else {
				// _logger.trace("zu wenige votes zum erkunden("+currentUnknownSwitch.getVotings().getVotingCount()+") fuer Switch bei "+currentUnknownSwitch);
			}
		}

		// pruefen, ob ein Switch angefragt wurde
		Switch[] switches = data.getSwitches().valuesAsArray(Switch.class);
		for (Switch currentSwitch : switches) {
			if (currentSwitch.getVotings().getVotingCount() >= MINIMUM_VOTERS) {
				if (currentSwitch.getSwitchingAgent() != "") {
					// switcher behaelt sein goal
					uncheckedAgents.remove(data.getAgent(currentSwitch
							.getSwitchingAgent()));
					continue;
				}
				// one of the voters will be switcher
				boolean foundSwitcher = false;
				List<Agent> votingAgents = currentSwitch.getVotings()
						.getVotingAgents();
				for (Agent possibleSwitcher : votingAgents) {
					// check, whether this voter can reach a switching Position
					// TODO jetzt wird es der erste in der Liste, welcher
					// erreichen kann
					// der am naechsten dran waer besser, 
					// wenn der nicht erreicht, is der zweite dran usw.
					List<Position> switchingPositions = currentSwitch.getSwitchingPositions(data.getWorld());
					int min = 999999;
					Position resultpos = null;
					for (Position openingPosition : switchingPositions) {
						IPathFindingHandle pathfinder = new PathFindingElement(false);
						PathResult pr = pathfinder.getPath(openingPosition,
								possibleSwitcher.getPosition(), data.getWorld(), false);
						if (pr.getErrortype() == null) {
							int length = pr.getPathLength();
							if (length < min) {
								resultpos = openingPosition;
								min = length;
							}
						}
					}
					if (resultpos == null) {
						// this agent cannot reach the switch, check the next
						// one
						continue;
					} else {
						foundSwitcher = true;
						// this agent can be switcher, assign goal
						rounds = getRoundsForSwitchGoal(currentSwitch);
						this.setNewGoal(possibleSwitcher, new GoalGoToSwitchingPosition(resultpos, currentSwitch, rounds));
						uncheckedAgents.remove(possibleSwitcher);
						currentSwitch.setSwitchingAgent(possibleSwitcher.getAddress().toString());
						// and do not check the other ones
						break;
					}
				}
				if (!foundSwitcher) {
					// if no voter can reach, send switch request to every other agent 
					Collection<Agent> allAgents = data.getAgents().values();
					//remove voters
					allAgents.removeAll(votingAgents);
					for (Agent a : allAgents) {
						sendSwitchRequestToAgent(currentSwitch, a);
					}
				}
			}
		}
	
	//-----------------HERDS------------------------------------------------
		//check update
		if (updateHerdsCounter < NUM_ROUNDS_BETWEEN_HERD_UPDATES) {
			updateHerdsCounter++;
			AgentMonitor.getInstance().updateHerds(herdOrga.getHerds());
		} else {
			//update the herd (cows, driving positions, distanceToCorral etc)
			herdOrga.updateHerds(NUM_ROUNDS_BETWEEN_HERD_UPDATES);
			updateHerdsCounter = 0;
			//to paint the new herds and driving positions
			AgentMonitor.getInstance().updateHerds(herdOrga.getHerds());
		}
		//check rebuild
		//if we don't have any herds rebuild faster!
		if(herdOrga.getHerds().size() == 0 && rebuildHerdsCounter < (NUM_ROUNDS_BETWEEN_HERD_REBUILDS - NUM_ROUNDS_TO_FASTER_REBUILD))
			rebuildHerdsCounter = NUM_ROUNDS_BETWEEN_HERD_REBUILDS - NUM_ROUNDS_TO_FASTER_REBUILD;
		if (rebuildHerdsCounter < NUM_ROUNDS_BETWEEN_HERD_REBUILDS) {
			rebuildHerdsCounter++;
		} else {
			//rebuild herds
			herdOrga.rebuildHerds();
			rebuildHerdsCounter = 0;
			AgentMonitor.getInstance().updateHerds(herdOrga.getHerds());
		}
		
		//check for all herds 
		//1. did they reached the corral? -> drivers can get a new goal
		//2. are there any fences in their way? -> vote for switch
		//3. are there enough drivers? -> don't need more drivers
		//4. if not rebuild update the driving position if the herds were updated
		// then delete all drivers from uncheckedAgents so that they won't get a new goal
		//5. find nearest drivers for remaining herds
		
		ArrayList<Herd> herds = new ArrayList<Herd>();
		herds.addAll(herdOrga.getHerds());
		ArrayList<Herd> herdsToRemove = new ArrayList<Herd>();
				
		if (!herds.isEmpty()) {
			for(Herd h: herds){
				//1. did they reach the corral? -> drivers can get a new goal
				if(h.insideCorral()){
					herdsToRemove.add(h);
					continue;
				}
				// 2. are there any fences in their way? -> switching!
				// TODO ueberpruefen und fuer mehrere fences voten!
				List<Point> fences;
				fences = h.getFences();
				// only herds with drivers should vote
				if (fences != null && !fences.isEmpty() && h.getDriverCount() > 0) {
					Point fence = fences.remove(0);
					Position s = data.getWorld().findSwitchForFence(new Position(fence.x, fence.y));
					MessageSwitchVote vote;
					if (s == null) {
						// there is no switch found yet
						vote = new MessageSwitchVote(fence.x, fence.y, true, h.getStepsToNextFence());
					} else {
						// vote for this switch to be opened
						vote = new MessageSwitchVote(s.getX(), s.getY(), false,	h.getStepsToNextFence());
					}
					this.handleSwitchVote(vote, h);
				}

				// 3. are there enough drivers? -> don't need more drivers
				if (h.getDriverCount() >= h.getMaxDriverCount()) {
					herdsToRemove.add(h);
				}

				// 4. if not rebuild: update the driving position if the herds
				// were updated
				// then delete all drivers from uncheckedAgents so that they
				// won't get a new goal
				for (Agent a : h.getDrivers()) {
					if (rebuildHerdsCounter != 0) {
						if (updateHerdsCounter == 0) {
							// update driving position
							this.setNewGoal(a, 
									new GoalDriveHerd(h.getMyDrivingPos(a.getAddress().toString()), h.getUpperLeft(), h.getLowerRight()));
						}
						// delete driver
						uncheckedAgents.remove(a);
					}
				}
			}
			herds.removeAll(herdsToRemove);
					
			//5. find nearest drivers for remaining herds
			Agent possibleDriver = null;
			Herd herd = null;
			List<Agent> uncheckedDrivers = new ArrayList<Agent>();
			uncheckedDrivers.addAll(uncheckedAgents);
			
			while (!herds.isEmpty() && !uncheckedDrivers.isEmpty()) {
				herd = herds.get(0);
				//sort uncheckedDrivers after their distance to the herd
				//therefor save the pathlength in the agent
				ArrayList<Agent> agentsToRemove = new ArrayList<Agent>();
				for(Agent a : uncheckedDrivers){
					a.setPathlengthToAHerd(herd.getPathLengthAsDistance(a.getPosition()));
					if(a.getPathlengthToAHerd() == -1){
						agentsToRemove.add(a);
					}
				}
				//delete all Agents that can't reach the herd
				uncheckedDrivers.removeAll(agentsToRemove);
				Collections.sort(uncheckedDrivers);
				if(uncheckedDrivers.isEmpty()){
					//no driver can reach the herd
					herds.remove(herd);
					uncheckedDrivers.clear();
					uncheckedDrivers.addAll(uncheckedAgents);
					continue;
				}
				//make the nearest Agents to Drivers of this herd 
				//leave MIN_EXPLORER Agents for exploring //TODO so wechseln die explorer dauernd -sinnvoll?
				while(uncheckedDrivers.size() + agentsToRemove.size() > MIN_EXPLORER && herd.getDriverCount() < herd.getMaxDriverCount()){
					possibleDriver = uncheckedDrivers.remove(0);
					if(herd.addDriver(possibleDriver)) {
						this.setNewGoal(possibleDriver, new GoalDriveHerd(herd.getMyDrivingPos(possibleDriver.getAddress().toString()), herd.getUpperLeft(), herd.getLowerRight()));
						uncheckedAgents.remove(possibleDriver);
					}
				}
				//prepare for next herd
				herds.remove(herd);
				uncheckedDrivers.clear();
				uncheckedDrivers.addAll(uncheckedAgents);
		}
	}
			
	//------------------EXPLORE------------------------------------------
		
		// verbleibende Agenten erkunden schicken
		IntHashMap<ExploringPoint> points = data.getPointsToExplore();
		List<Agent> exploringAgents = new LinkedList<Agent>();
		for (Agent agent : uncheckedAgents) {
			if (!points.isEmpty()) {
				// find exploring point next to agent
				ExploringPoint targetp = null;
				int min = 99999;
				int[] keysp = points.keysAsArray();
				for (int keyp : keysp) {
					ExploringPoint p = points.get(keyp);
					if (p.getAgent() == null) {// we need only one Agent running
												// to this point
						int minStepsForThisPoint = data.getWorld().distance(p,
								agent.getPosition());
						if (minStepsForThisPoint < min) {
							min = minStepsForThisPoint;
							targetp = p;
						}
					}
				}
				if (targetp != null) {
					// store this goal in the agent
					targetp.setAgent(agent.getAddress().toString());
					AbstractGoal goal = new GoalExploreField(targetp);
					this.setNewGoal(agent, goal);
					exploringAgents.add(agent);
				}
			}
		}
		uncheckedAgents.removeAll(exploringAgents);
		
		//TODO hier verbleibende Agenten irgendwas sinnvoles tun lassen... treiben helfen, gegnerisches corral stuermen usw
		if(data.isFoundEnemyCorral()){
			// send agents to a random position in enemy corral
			Random rn = new Random();
			int corralwidth = data.getEnemyCorralLowerRight().getX() - data.getEnemyCorralUpperLeft().getX();
			int corralheight = data.getEnemyCorralLowerRight().getY() - data.getEnemyCorralUpperLeft().getY();
			Position corralpoint;
			int x = 0;
			int y = 0;
			for(Agent a : uncheckedAgents){
				if(corralwidth != 0)
					x = rn.nextInt(corralwidth);
				if(corralheight != 0)
					y = rn.nextInt(corralheight);
				corralpoint = new Position(data.getEnemyCorralUpperLeft().getX() + x, data.getEnemyCorralUpperLeft().getY() + y);
				this.setNewGoal(a, new GoalBlockEnemyCorral(corralpoint));
			}
		}
	}

	private int getRoundsForSwitchGoal(Switch s) {
		return s.getVotings().getMaxDistance();
	}

	/**
	 * sends a message containing the current goal to the agent<br/>
	 * the goal has to be stored in the agent
	 * 
	 * @param agent the agent which should receive his goal
	 */
	private void sendGoalToAgent(Agent agent) {
		if (agent.getGoal() != null) {
			// send goal
			IMessage message = _communicationHandle.createMessage();
			message.setContent(agent.getGoal());
			try {
				_communicationHandle.sendMessage(agent.getAddress(), message);
			} catch (Exception e) {
				_logger.error("Could not send goal", e);
			}
		}
	}

	/**
	 * sends a switch request-message to an agent to ask, whether he can reach the switch
	 *  
	 * @param s the switch to open
	 * @param agent the agent to ask
	 */
	private void sendSwitchRequestToAgent(Switch s, Agent agent) {

		s.addRequestedAgent(agent.getAddress().toString());

		IMessage message = _communicationHandle.createMessage();
		message.setContent(new SwitchAnfrage(s.getX(), s.getY()));
		_logger.trace("COORD: sending SWITCH REQUEST to agent: " + s);
		try {
			_communicationHandle.sendMessage(agent.getAddress(), message);
		} catch (Exception e) {
			_logger.error("Could not send Switchrequest to agent", e);
		}
	}

	/**
	 * initializes new datacontainers<br/>
	 * these are:<br/>
	 * - data<br/>
	 * - herdOrga
	 * 
	 * @param world the new Gridworld
	 */
	private void initializeData(Gridworld world) {
		data = new CoordinationData(world);
		herdOrga = new HerdOrganisation(world);
	}

	/**
	 * sets a new goal to an agent, but before this, the agent is 
	 * removed from things referring to the old goal, e.g.
	 * switches or exploringpoints
	 * @param a the agent
	 * @param g the new goal
	 */
	private void setNewGoal(Agent a, AbstractGoal g) {
		AbstractGoal oldGoal = a.getGoal(); 
		boolean remove=true;
		if(oldGoal != null && oldGoal instanceof GoalDriveHerd){
			Point upperLeft = ((GoalDriveHerd) oldGoal).getUpperLeft();
			Point lowerRight = ((GoalDriveHerd) oldGoal).getLowerRight();
			if(g!=null && g instanceof GoalDriveHerd){
				Point newUpperLeft = ((GoalDriveHerd) oldGoal).getUpperLeft();
				Point newLowerRight = ((GoalDriveHerd) oldGoal).getLowerRight();
				if((upperLeft.x==newUpperLeft.x) && (upperLeft.y==newUpperLeft.y)&&(lowerRight.x==newLowerRight.x)&&(lowerRight.y==newLowerRight.y))
					//only an update
					remove=false;
			}
			if(remove)
				for(Herd h: herdOrga.getHerds()){
					if((h.getLowerRight().x == lowerRight.x) && (h.getLowerRight().y==lowerRight.y)&& (h.getUpperLeft().x == upperLeft.x) && (h.getUpperLeft().y==upperLeft.y)){
						_logger.trace("setNewGoal: remove agent from: "+h.toString());
						h.removeDriver(a);
						break;
					}
				}
		}
		data.setNewGoal(a, g);
	}

	/**
	 * updates the own worldmodel with the GridFields seen by an agent
	 * @param p the perception with new information about the world
	 * @param sender the Agent who sent this perception
	 */
	private void updateData(Perception p, IAddress sender) {
		data.update(p, sender);
	}

	@Override
	public void start() {
		super.start();

		// Referenz auf den Agenten
		IAgent agent = AgentScope.getAgentReference();
		// Referenz auf die Kommunikationsschnittstelle
		_communicationHandle = (ICommunicationHandle) agent
				.getHandle(ICommunicationHandle.class);
		// Referenz auf den Logger
		_logger = agent.getLogger("Coordination");

		// define the group
		try {
			// use the teamname as groupname for communication
			String groupName = (String) agent
					.getProperty(IMAPCConstants.TEAMNAME_KEY);
			// get the address for this group
			_address = _communicationHandle
					.getMulticastAddressForName(groupName);
			// join the group
			_communicationHandle.joinGroup(_address);
			_communicationHandle.joinGroup(_communicationHandle
					.getMulticastAddressForName("coordinator"));
			// sendCoordinatorResponse();
		} catch (IOException e) {
			_logger.error("Could not join the team group", e);
		}
	}

	@Override
	public void stop() {
		// Der Agent faehrt runter -> wir brauchen die Gruppe nicht mehr.
		try {
			_communicationHandle.leaveGroup(_address);
		} catch (IOException e) {
			_logger.error("Could not leave the team group", e);
		}

		super.stop();
	}
}
