package de.tub.mapc.knowledge;

import static de.tub.mapc.IMAPCConstants.TYPE_FENCE;
import static de.tub.mapc.IMAPCConstants.TYPE_OPEN_FENCE;
import static de.tub.mapc.IMAPCConstants.TYPE_TREE;
import static de.tub.mapc.IMAPCConstants.TYPE_UNKNOWN;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;

import de.tub.mapc.Coordination.SwitchVotes;
import de.tub.mapc.Coordination.Voter;
import de.tub.mapc.Coordination.messages.SwitchingAnswer;
import de.tub.mapc.worldmodel.GridField;

public class Switch extends Position {
	
	private List<Position> switchingPositions; 
	private String switchingAgent; 
	private SwitchVotes votings;
	private List<String> notReachingAgents;
	private PriorityQueue<SwitchingAnswer> possibleSwitchers;
	private List<String> requestedAgents;
	private int successfullCheckedDirections;
	private CompareSwitchingAnswer comp;
//	private int timeOfLastAnswer=0;

	public Switch() {
		super();
		init();
	}

	public Switch(int x, int y) {
		super(x, y);
		init();
	}
	
	private void init(){
		switchingPositions = new ArrayList<Position>();
		votings = new SwitchVotes();
		notReachingAgents = new ArrayList<String>();
		requestedAgents = new ArrayList<String>();
		comp = new CompareSwitchingAnswer();
		possibleSwitchers = new PriorityQueue<SwitchingAnswer>(10, comp);
		successfullCheckedDirections = 0;
		switchingAgent = "";
//		timeOfLastAnswer=0;
	}
	
	public void addVoter(Voter voter, int dist){
		votings.addVoter(voter, dist);
	}
	
	public void addPossibleSwitcher(SwitchingAnswer sa){
		possibleSwitchers.add(sa);
	}
	
	public void addnonReachingAgent(String agent){
		if(!notReachingAgents.contains(agent)){
			notReachingAgents.add(agent);
		}
	}
	
	public SwitchVotes getVotings(){
		return votings;
	}
	
	public List<String> getNotReachingAgents() {
		return notReachingAgents;
	}

	/*public void resetForNewSwitcher(){
		notReachingAgents = new ArrayList<String>();
		possibleSwitchers = new HashMap<String, Integer>();
		requestedAgents = new ArrayList<String>();
		switchingAgent = "";
	}*/
	public void removePossibleSwitcher(String agent){
		possibleSwitchers.remove(agent);
	}
	
//	public void setTime(int t){
//		this.timeOfLastAnswer = t;
//	}
	
//	public int getTime(){
//		return timeOfLastAnswer;
//	}

	public void clear(){
		votings = new SwitchVotes();
		notReachingAgents = new ArrayList<String>();
		possibleSwitchers = new PriorityQueue<SwitchingAnswer>(10, comp);
		requestedAgents = new ArrayList<String>();
		switchingAgent = "";
//		timeOfLastAnswer = 0;
		
	}
	
	/**
	 * stores the positions where the agent can go to use the switch
	 * 
	 * @param world
	 * @return Goal
	 */
	private void computeSwitchingPositions(Gridworld world) {
		GridField[][] grid = world.getGrid();
		GridField currentNeighbor;
		try {
			currentNeighbor = grid[getX() - 1][getY()];
			if(currentNeighbor.getType() != TYPE_UNKNOWN){
				successfullCheckedDirections++;
			}
			if ((currentNeighbor.getType() != TYPE_FENCE)
					&& (currentNeighbor.getType() != TYPE_OPEN_FENCE)
					&& (currentNeighbor.getType() != TYPE_TREE)) {
				if (!switchingPositions.contains(currentNeighbor))
					switchingPositions.add(currentNeighbor);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			// nothing, switch war direkt am rand der Karte
			successfullCheckedDirections++;
		}
		try {
			currentNeighbor = grid[getX() + 1][getY()];
			if(currentNeighbor.getType() != TYPE_UNKNOWN){
				successfullCheckedDirections++;
			}
			if ((currentNeighbor.getType() != TYPE_FENCE)
					&& (currentNeighbor.getType() != TYPE_OPEN_FENCE)
					&& (currentNeighbor.getType() != TYPE_TREE)) {
				if (!switchingPositions.contains(currentNeighbor))
					switchingPositions.add(currentNeighbor);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			// nothing
			successfullCheckedDirections++;
		}
		try {
			currentNeighbor = grid[getX()][getY() - 1];
			if(currentNeighbor.getType() != TYPE_UNKNOWN){
				successfullCheckedDirections++;
			}
			if ((currentNeighbor.getType() != TYPE_FENCE)
					&& (currentNeighbor.getType() != TYPE_OPEN_FENCE)
					&& (currentNeighbor.getType() != TYPE_TREE)) {
				if (!switchingPositions.contains(currentNeighbor))
					switchingPositions.add(currentNeighbor);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			// nothing
			successfullCheckedDirections++;
		}
		try {
			currentNeighbor = grid[getX()][getY() + 1];
			if(currentNeighbor.getType() != TYPE_UNKNOWN){
				successfullCheckedDirections++;
			}
			if ((currentNeighbor.getType() != TYPE_FENCE)
					&& (currentNeighbor.getType() != TYPE_OPEN_FENCE)
					&& (currentNeighbor.getType() != TYPE_TREE)) {
				if (!switchingPositions.contains(currentNeighbor))
					switchingPositions.add(currentNeighbor);
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			// nothing
			successfullCheckedDirections++;
		}
	}

	public List<Position> getSwitchingPositions(Gridworld world) {
		if (successfullCheckedDirections < 4)computeSwitchingPositions(world);
		return switchingPositions;
	}

	public void setSwitchingAgent(String switchingAgent) {
		this.switchingAgent = switchingAgent;
	}

	public String getSwitchingAgent() {
		return switchingAgent;
	}

	public void addRequestedAgent(String agent) {
		if(!requestedAgents.contains(agent)){
			requestedAgents.add(agent);
		}
	}

	public List<String> getRequestedAgents() {
		return requestedAgents;
	}

	public PriorityQueue<SwitchingAnswer> getPossibleSwitchers() {
		return possibleSwitchers;
	}

	
	private class CompareSwitchingAnswer implements Comparator<SwitchingAnswer>{
		
		@Override
		public int compare(SwitchingAnswer a, SwitchingAnswer b){
			//a.calculateCost();b.calculateCost();
			if (a.getDistance() == b.getDistance()) return 0;
			if (a.getDistance() > b.getDistance()) return 1;
			return -1;
		}
		
	}

//	@Override
//	public void deserialise(IDeserialiser arg0) throws IOException,
//			ClassNotFoundException {
//		super.deserialise(arg0);
//		int size = arg0.readInt();
//		for (int i = 0; i < size; i++) {
//			Position p = (Position) arg0.readObject();
//			switchingPositions.add(p);
//		}
//		switchingAgent = arg0.readString();
//		size = arg0.readInt();
//		for (int i = 0; i < size; i++) {
//			String s = arg0.readString();
//			votings.add(s);
//		}
//		size = arg0.readInt();
//		for (int i = 0; i < size; i++) {
//			String s = arg0.readString();
//			notReachingAgents.add(s);
//		}
//		size = arg0.readInt();
//		for (int i = 0; i < size; i++) {
//			String s = arg0.readString();
//			requestedAgents.add(s);
//		}
//		size = arg0.readInt();
//		for (int i = 0; i < size; i++) {
//			String s = arg0.readString();
//			Integer d = Integer.valueOf(arg0.readInt());
//			possibleSwitchers.put(s,d);
//		}
//		successfullCheckedDirections = arg0.readInt();
//
//	}
//
//	@Override
//	public void serialise(ISerialiser arg0) throws IOException {
//		super.serialise(arg0);
//		arg0.writeInt(switchingPositions.size());
//		for (Position p : switchingPositions) {
//			arg0.writeObject(p);
//		}
//		arg0.writeString(switchingAgent);
//		arg0.writeInt(votings.size());
//		for (String s: votings) {
//			arg0.writeString(s);
//		}
//		arg0.writeInt(notReachingAgents.size());
//		for (String s: notReachingAgents) {
//			arg0.writeString(s);
//		}
//		arg0.writeInt(requestedAgents.size());
//		for (String s: requestedAgents) {
//			arg0.writeString(s);
//		}
//		arg0.writeInt(possibleSwitchers.size());
//		Set<String> keys = possibleSwitchers.keySet();
//		for (String s: keys) {
//			Integer d = possibleSwitchers.get(s); 
//			arg0.writeString(s);
//			arg0.writeInt(d.intValue());
//		}
//		arg0.writeInt(successfullCheckedDirections);
//	}

	@Override
	public String toString() {
		return "Switch (" + getX() + ", " + getY() + ")";
	}

}
