package de.tub.mapc.monitor;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.util.List;

import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.Coordination.Agent;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.worldmodel.GridField;
import de.tub.mapc.worldmodel.Herd;

public class MapComponent extends Component {
	/**
	 * 
	 */
	private static final long serialVersionUID = -4965341891553823554L;
	
    public final static Color[] COLORS;
    
    public final static Color[] HERD_COLORS;
    
    static {
        COLORS= new Color[IMAPCConstants.TYPE_COUNT];
        COLORS[IMAPCConstants.TYPE_UNKNOWN]= Color.LIGHT_GRAY;
        COLORS[IMAPCConstants.TYPE_EMPTY]= Color.GREEN;
        COLORS[IMAPCConstants.TYPE_CORRAL]= Color.ORANGE;
        COLORS[IMAPCConstants.TYPE_ENEMY_CORRAL]= Color.CYAN;
        COLORS[IMAPCConstants.TYPE_PARTNER]= Color.RED;
        COLORS[IMAPCConstants.TYPE_ENEMY]= Color.BLUE;
        COLORS[IMAPCConstants.TYPE_COW]= Color.WHITE;
        COLORS[IMAPCConstants.TYPE_FENCE]= Color.BLACK;
        COLORS[IMAPCConstants.TYPE_OPEN_FENCE]= Color.GRAY;
        COLORS[IMAPCConstants.TYPE_SWITCH]= Color.PINK;
        COLORS[IMAPCConstants.TYPE_TREE]= new Color(034, 139, 034);
        
        HERD_COLORS= new Color[]{
            new Color(255,119,198),
            new Color(253,0,140),
            new Color(149,0,74),
            new Color(132,17,100),
            new Color(172,92,243),
            new Color(105,14,166),
            new Color(149,0,140)
        };
    }
	
	
	private GridField[][] world;
//	private static int IMGSIZE = 10;
//	private int selectedAgentX;
//	private int selectedAgentY;
	private int selectedGoalX;
	private int selectedGoalY;
	private int currStep = 0;
	private List<Herd> herds;

	public void setHerds(List<Herd> h) {
		herds = h;
		repaint();
	}
	
	public void setCurrStep(int step){
		this.currStep=step;
	}

//	public void setSelectedPosition(int x, int y) {
//		selectedAgentX = x;
//		selectedAgentY = y;
//	}

	public void setSelectedGoal(int x, int y) {
		selectedGoalX = x;
		selectedGoalY = y;
	}

	public void setSelectedGoal(Position p) {
		if(p != null){
			selectedGoalX = p.getX();
			selectedGoalY = p.getY();
		}
	}

	public MapComponent() {
		super();
	}

	public void setWorld(GridField[][] world) {
		this.world = world;
		repaint();
	}

	@Override
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;
		
		final int imgsize= Math.max(2, Math.min(getWidth() / world.length, getHeight() / world[0].length));

		// paint the gridworld
		for (int x = 0; x < world.length; x++) {
			for (int y = 0; y < world[0].length; y++) {
				if (world[x][y] != null) {
					int type = world[x][y].getType();
                    g2.setColor(COLORS[type]);
					//fog of war
					if(currStep-world[x][y].getSeenInStep()>3 && world[x][y].getType()!= IMAPCConstants.TYPE_UNKNOWN)
						g2.setColor(new Color(g2.getColor().getRed(),g2.getColor().getGreen(), g2.getColor().getBlue(), 120));
					g2.fillRect(imgsize * x, imgsize * y, imgsize - 1,
							imgsize - 1);
				}
			}
		}// end paint the gridworld

//		// draw the selected agent
//		g2.setColor(Color.YELLOW);
//		g2.fillRect(IMGSIZE * selectedAgentX, IMGSIZE * selectedAgentY, IMGSIZE - 1, IMGSIZE - 1);

		// draw the goal of the selected agent
		g2.setColor(Color.YELLOW);
		g2.fillRect(imgsize * selectedGoalX, imgsize * selectedGoalY, imgsize - 1, imgsize - 1);

		// draw the herds
		List<Position> drivingPos;
		List<Agent> drivers;
		if(herds != null)
			for(int i = 0; i < herds.size(); i++){
				//paint each herd in a different color
			    if(i < HERD_COLORS.length) {
			        g2.setColor(HERD_COLORS[i]);
			    } else {
			        g2.setColor(Color.BLACK);
			    }
			    
				// draw Herd
				int upperleftx = imgsize * herds.get(i).getUpperLeft().x;
				int upperlefty = imgsize * herds.get(i).getUpperLeft().y;
				int herdwidth = herds.get(i).getLowerRight().x - herds.get(i).getUpperLeft().x+1;
				int herdheight = herds.get(i).getLowerRight().y - herds.get(i).getUpperLeft().y+1;
				g2.drawRect(upperleftx, upperlefty, imgsize * herdwidth, imgsize * herdheight);
				//draw driving positions
				drivingPos = herds.get(i).getDrivingPositions();
				if(drivingPos != null)
					for(Position p : drivingPos){
						g2.drawOval(imgsize * p.getX(), imgsize * p.getY(), imgsize - 1, imgsize - 1);
					}
				//draw drivers
				drivers = herds.get(i).getDrivers();
				if(drivers != null)
					for(Agent a : drivers){
						g2.fillRect(imgsize * a.getPosition().getX(), imgsize * a.getPosition().getY(), imgsize - 1, imgsize - 1);
					}
			}
	}
}
