package de.tub.mapc.pathfinding;

import java.awt.Point;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import de.dailab.keks.io.IDeserialiser;
import de.dailab.keks.io.ISerialiser;
import de.jiac.micro.agent.memory.IFact;
import de.tub.mapc.knowledge.Position;

/**
 * 
 * contains some information about the pathfinding 
 * in this step
 * 
 * first, ask for errortype and reachedgoal
 * 
 * @author Erik
 *
 */
public class PathResult implements IFact{
	
	private int pathLength = -1;

	/**
	 * an action for this step, may be null, if there was an error or the goal is reached
	 */
	private String thisStep;	

	/**
	 * describes the type of the error if there was one<br/>
	 * Constants defined in IMAPConstants
	 */
	private String errortype;	
	
	/**
	 * number of steps, until the path crosses a fence<br/>
	 * -1 if there is no fence 
	 */
	private int stepsToNextFence;
	
	/**
	 * x-position of the next fence in path<br/>
	 * (only if there is a fence)
	 */
	private int fencex;
	
	/**
	 * y-position of the next fence in path<br/>
	 * (only if there is a fence)
	 */
	private int fencey;
	
	/**
	 * list with all the fences in the path<br/>
	 * represented by Point classes
	 */
	private List<Point> fences;
	
	/**
	 * flag shows, if we are on the destination
	 */
	private boolean reachedGoal;
	
	private Position threeStepsPosition;
	
	/**
	 * creates a new PathResult with default values
	 */
	public PathResult(){
		stepsToNextFence = -1;
		fencex = -1;
		fencey = -1;
		fences = new ArrayList<Point>();
		setReachedGoal(false);
	}
	
	public PathResult(String action){
		stepsToNextFence = -1;
		fencex = -1;
		fencey = -1;
		fences = new ArrayList<Point>();
		setReachedGoal(false);
		setThisStep(action);
	}

	public int getPathLength() {
		return pathLength;
	}

	public void setPathLength(int pathLength) {
		this.pathLength = pathLength;
	}

	/**
	 * an action for this step, may be null, if there was an error or the goal is reached
	 */
	public String getThisStep() {
		return thisStep;
	}

	/**
	 * an action for this step, may be null, if there was an error or the goal is reached
	 */
	public void setThisStep(String thisStep) {
		this.thisStep = thisStep;
	}

	/**
	 * describes the type of the error if there was one<br/>
	 * Constants defined in IMAPConstants
	 */
	public String getErrortype() {
		return errortype;
	}

	/**
	 * describes the type of the error if there was one<br/>
	 * Constants defined in IMAPConstants
	 */
	public void setErrortype(String errortype) {
		this.errortype = errortype;
	}

	/**
	 * number of steps, until the path crosses a fence<br/>
	 * -1 if there is no fence 
	 */
	public int getStepsToNextFence() {
		return stepsToNextFence;
	}

	/**
	 * number of steps, until the path crosses a fence<br/>
	 * -1 if there is no fence 
	 */
	public void setStepsToNextFence(int stepsToNextFence) {
		this.stepsToNextFence = stepsToNextFence;
	}

	/**
	 * x-position of the next fence in path<br/>
	 * (only if there is a fence)
	 */
	public void setFencex(int fencex) {
		this.fencex = fencex;
	}

	/**
	 * x-position of the next fence in path<br/>
	 * (only if there is a fence)
	 */
	public int getFencex() {
		return fencex;
	}

	/**
	 * y-position of the next fence in path<br/>
	 * (only if there is a fence)
	 */
	public void setFencey(int fencey) {
		this.fencey = fencey;
	}

	/**
	 * y-position of the next fence in path<br/>
	 * (only if there is a fence)
	 */
	public int getFencey() {
		return fencey;
	}

	/**
	 * list with all the fences in the path<br/>
	 * represented by Point classes
	 */
	public List<Point> getFences() {
		return fences;
	}
	
	/**
	 * list with all the fences in the path<br/>
	 * represented by Point classes
	 */
	public List<Point> setFences() {
		return fences;
	}
	
	/**
	 * add a fence to the list
	 */
	public void addFence(Point fence) {
		fences.add(fence);
	}
	
	/**
	 * flag shows, if we are on the destination
	 */
	public void setReachedGoal(boolean reachedGoal) {
		this.reachedGoal = reachedGoal;
	}

	/**
	 * flag shows, if we are on the destination
	 */
	public boolean isReachedGoal() {
		return reachedGoal;
	}

	public void setThreeStepsPosition(Position n) {
		threeStepsPosition = n;
		
	}
	
	/**
	 * the position of the field in 3 steps
	 * 
	 * @return
	 */
	public Position getThreeStepsPosition() {
		return threeStepsPosition;
		
	}
	
	@Override
	public void deserialise(IDeserialiser arg0) throws IOException,
			ClassNotFoundException {
		this.thisStep = arg0.readString();
		this.errortype= arg0.readString();
		this.stepsToNextFence = arg0.readInt();
		this.fencex = arg0.readInt();
		this.fencey = arg0.readInt();
		this.reachedGoal = arg0.readBoolean();
		this.threeStepsPosition = (Position) arg0.readObject();
		
	}

	@Override
	public void serialise(ISerialiser arg0) throws IOException {
		arg0.writeString(thisStep);
		arg0.writeString(errortype);
		arg0.writeInt(stepsToNextFence);
		arg0.writeInt(fencex);
		arg0.writeInt(fencey);
		arg0.writeBoolean(reachedGoal);
		arg0.writeObject(threeStepsPosition);
	}

	
	
	
	
}
