package de.tub.mapc.worldmodel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;


import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.Coordination.Agent;
import de.tub.mapc.knowledge.Gridworld;

public class HerdOrganisation {

	/**
	 * list of herds
	 */
	private List<Herd> herds = new ArrayList<Herd>();
	
	/**
	 * gridworld
	 */
	private Gridworld world;
		
	final static int MINIMUM_HERD_SIZE = 1; //TODO was wenn nachher nur noch einzelne kuehe rumlaufen?
	final static int MAX_HERD_SIZE = 25;
	final static int MAX_DIST_BETWEEN_COWS_IN_HERD = 3; //this seems to be the best number
														
	
	
	/**
	 * @param w Gridworld
	 */
	public HerdOrganisation(Gridworld w){		
		world = w;	
	}
	
	public List<Herd> getHerds(){
		Collections.sort(herds);
		return herds;
	}
	
	/**
	 * updates the world of all herds and the herdOrganisation
	 * @param w the new Gridworld
	 */
	public void updateWorldOfHerds(Gridworld w){
		world = w;
		if(herds.size() != 0)
			for(Herd h : herds){
				h.updateWorld(w);
			}
	}
	
	/**
	 * updates the herds so that they follow their cows ;-)<br/>
	 * 1. updates every single herd<br/>
	 * 2. merges overlapping herds<br/>
	 * 3. trims the herds
	 * @param stepsBetweenUpdates the number of steps, where no update has been performed
	 */
	public void updateHerds(int stepsBetweenUpdates){
		
		// update every single herd
		if(herds.size() != 0)
			for(Herd h : herds){
				h.updateHerd(stepsBetweenUpdates);
			}
		// then merge overlapping herds
		if(herds.size() > 1){
			herds = mergeHerds(herds);
			herds = trimHerds(herds); 
			//TODO richtig so? geht das dann noch mit der treiberzuordnung?
			// scheint zu klappen, geht auf jeden fall besser so, weil sich
			// sonst vor dem eingang zum gatter eine riesige herde bildet die
			// immer neu ankommende hinzufuegt und somit die treibepositionen
			// extrem nach aussen verlagert -- Erik
		}		
	}
	
	/**
	 * merges the herds in the given list by putting overlapping herds into one new
	 * 
	 * @param herds the list with the old herds
	 * @return a List with new Herds
	 */
	private List<Herd> mergeHerds(List<Herd> herds) {
		ArrayList<Herd> newHerds = new ArrayList<Herd>();
		newHerds.addAll(herds);
		Herd h1;
		Herd h2;
		Herd h3;
		for (int i = 0; i < newHerds.size(); i++) {
			h1 = newHerds.get(i);
			for (int index = 0; index < newHerds.size(); index++) {
				h2 = newHerds.get(index);
				if (overlap(h1, h2)) {
					h3 = merge(h1, h2);
					newHerds.add(h3);
					newHerds.remove(h1);
					newHerds.remove(h2);
				}
			}
		}
		return newHerds;
	}
	
	/**
	 * combines the cows of two herds to a new Herd
	 * 
	 * @param h1 the first Herd
	 * @param h2 the second Herd
	 * @return a new Herd containing the cows of h1 and h2
	 */
	private Herd merge(Herd h1, Herd h2) {
		ArrayList<GridField> cows = new ArrayList<GridField>();
		cows.addAll(h1.getCows());
		for(GridField cow : h2.getCows()){
			if(!cows.contains(cow)) cows.add(cow);
		}
		return new Herd(cows, world);
	}

	/**
	 * checks, whether 2 herds are overlapping
	 * 
	 * @param h1 the first herd
	 * @param h2 the second herd
	 * @return true, if there is at least one field belonging to both herds
	 */
	private boolean overlap(Herd h1, Herd h2) {
		
		// Bestimmen der Eckpunktkoordinaten des beide Herden umschreibenden Rechtecks
		int outerRectangleLeftBoundary = Math.min(h1.getUpperLeft().x, h2.getUpperLeft().x);
		int outerRectangleRightBoundary = Math.max(h1.getLowerRight().x, h2.getLowerRight().x);
		int outerRectangleUpperBoundary = Math.min(h1.getUpperLeft().y, h2.getUpperLeft().y);
		int outerRectangleLowerBoundary = Math.max(h1.getLowerRight().y, h2.getLowerRight().y);

		int outerRectangleWidth = outerRectangleRightBoundary - outerRectangleLeftBoundary; 
		int outerRectangleHeight = outerRectangleLowerBoundary - outerRectangleUpperBoundary;
		
		int h1width = h1.getLowerRight().x - h1.getUpperLeft().x;
		int h2width = h2.getLowerRight().x - h2.getUpperLeft().x;
		int h1height = h1.getLowerRight().y - h1.getUpperLeft().y; 
		int h2height = h2.getLowerRight().y - h2.getUpperLeft().y;
		
		// Pruefen auf Kollision (umschreibendes Rechteck kleiner als Summe der Herden)
		
		// TODO hier reicht doch ein || statt &&? wenn sie breiter sind dann ueberlappen sie sich doch auch...
		// ich aender das jetzt mal, wenns falsch ist kann es zurueckgeaendert werden. aber sonst kann es
		// gleichhohe Herden geben, die sich in der Breite ueberlappen (kam bei mir oefter vor) -- Erik
		
		// hmm ok mit || funktioniert es nicht, da werden die Herden riesig. versteh aber nicht waum :( -- Erik
		if((h1width + h2width > outerRectangleWidth) && (h1height + h2height > outerRectangleHeight))
				return true;
		
		return false;
	}

	/**
	 * rebuilds the herds<br/>
	 * 
	 * 1. every cow is searched in the grid<br/>
	 * 2. the cows are related to herds<br/>
	 * 3. overlapping herds are merged 
	 */
	public void rebuildHerds(){
		// find all cows in grid
		List<GridField> cows = new ArrayList<GridField>();
		for (GridField[] gf : world.getGrid()) {
			for (GridField f : gf) {
				if (f.getType() == IMAPCConstants.TYPE_COW) {
					cows.add(f);
				}
			}
		}
	
		//we want to rebuild the herds, so we have to delete the old herds!
		herds.clear();

		// relate the cows to herds
		if(!cows.isEmpty()){
			List<Herd> newHerds = relateCowsToHerds(cows);
			// delete small herds
			for (Herd h : newHerds) {
				if(h.size() >= MINIMUM_HERD_SIZE) 
					herds.add(h);
			}
			// merge overlapping herds
			if(herds.size() > 1){
				herds = mergeHerds(herds);				
			}	
		}
	}
	
	/**
	 * 
	 * @param cows
	 * @return
	 */
	private List<Herd> relateCowsToHerds(List<GridField> cows) {
		//TODO Cluster-Erkennung
		ArrayList<Herd> newHerds = new ArrayList<Herd>();
		ArrayList<GridField> herd = new ArrayList<GridField>();

		synchronized(cows) {
			while(!cows.isEmpty()){
				herd.add(cows.get(0));
				for(GridField f : cows) {
						int index = 0;
						while(index < herd.size() && herd.size() < MAX_HERD_SIZE) {
							if(Math.abs(f.getX() - herd.get(index).getX()) <= MAX_DIST_BETWEEN_COWS_IN_HERD &&
									  Math.abs(f.getY() - herd.get(index).getY()) <= MAX_DIST_BETWEEN_COWS_IN_HERD) {
								if(!herd.contains(f))
									herd.add(f);
		
							}					
							index++;
						}
				}
				cows.removeAll(herd);
				if(herd.size() > 0) 
					newHerds.add(new Herd(herd, world));
				herd.clear();
			}	
		}	
		return trimHerds(newHerds);
	}	
	
	/**
	 * trims the herds by diving big herds to smaller ones
	 * @param herds
	 * @return
	 */
	private List<Herd> trimHerds(List<Herd> herds) {
		
		// first trim herds to right form
		int index = 0;
		int sizeatbegin = herds.size(); // we dont need to go over the new created herds TODO is that correct? -- Erik 
		while(index < sizeatbegin) {
			Herd h = herds.remove(0);
			int herdheight = h.getLowerRight().y - h.getUpperLeft().y;
			int herdwidth = h.getLowerRight().x - h.getUpperLeft().x;
			if(2 * Math.sqrt(h.size()) + 1 < herdheight) {
				// herd stretches too far vertically - divide it
				ArrayList<GridField> upperCows = new ArrayList<GridField>();
				ArrayList<GridField> lowerCows = new ArrayList<GridField>();
				for(GridField f : h.getCows()) {
					if(f.getY() < h.getMidlle().getY()) {
						upperCows.add(f);
					} else {
						lowerCows.add(f);
					}					
				}
				herds.addAll(makeMeTwoNewHerds(upperCows, lowerCows, h));
			} else if (2*h.size()+1 < herdwidth) {
				// herd stretches too far horizontally - divide it
				ArrayList<GridField> leftCows = new ArrayList<GridField>();
				ArrayList<GridField> rightCows = new ArrayList<GridField>();
				for(GridField f : h.getCows()) {
					if(f.getX() < h.getMidlle().getX()) {
						leftCows.add(f);
					} else {
						rightCows.add(f);
					}					
				}
				herds.addAll(makeMeTwoNewHerds(leftCows, rightCows, h));
			} else {
				herds.add(h);
			}
			index++;
		}
		//then to right sizes TODO
		return herds;
	}
	
	/**
	 * creates two new herds from the cows of one old herd<br/>
	 * the cows have to be divided into  2 groups before calling his method
	 * @param cows1 the first group of cows
	 * @param cows2 the second group of cows
	 * @param h the old herd, drivers and world are tken from this one
	 * @return a list containing the two new herds
	 */
	private ArrayList<Herd> makeMeTwoNewHerds(ArrayList<GridField> cows1, ArrayList<GridField> cows2, Herd h) {
		ArrayList<Herd> herds = new ArrayList<Herd>();
		
		if(cows1.size() >= MINIMUM_HERD_SIZE){
			Herd herd1 = new Herd(cows1, h.getWorld());
			for(Agent a : h.getDrivers()) {
				herd1.addDriver(a);
			}
			herds.add(herd1);
		}
		if(cows2.size() >= MINIMUM_HERD_SIZE){
			Herd herd2 = new Herd(cows2, h.getWorld());
			herds.add(herd2);
		}
		return herds;
	}
	
}
